/*
-------------------------------------------------------------------
SUDOKU SENSEI 1.03: a Sudoku Explainer Engine
Copyright (C) 2005  Martin Knoblauch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
MA  02110-1301, USA.

Contact the author: comocomocomo AT users.sourceforge.net
Latest versions: http://sourceforge.net/projects/sudoku-sensei
-------------------------------------------------------------------

    demo4-creator.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "../../../sudoku-sensei/sudoku-pub.h"

#ifdef WINDOWS
#define SYSTEM_PAUSE system ("pause")
#else
#define SYSTEM_PAUSE
#endif

/*////////////////////////////////////////////////////////////// */

#define MAX_GRIDS 100
#define MAX_DIST 1000
#define MAX_NUM 10000
#define MAX_TRIALS 10

int my_random (int num);

typedef struct s_params
{
  int blockWidth, blockHeight;
  int nGrids;
  int * offsetsX, * offsetsY;
  char samurai, diagonals;
  int num;
  int restrictinters;
}
t_params;

int ParseParams (t_params * p, int argc, char *argv[]);
void FreeParams (t_params * p);

int CreateProblems (t_board_p b, t_params * p,
                    int * cells, int * symbols,
                    char * given,
                    int serial);

/*////////////////////////////////////////////////////////////// */

int main (int argc, char *argv[])
{
  t_params par;
  t_board_p b;
  int n, nCl, serial;
  int * cells, * symbols;
  char * given;
  FILE * pf;

  srand ((unsigned)time(NULL));

  PrintGPL (stdout);  /*Show License */

  if (ParseParams (&par, argc, argv) != 0)
  {
    SYSTEM_PAUSE;
    return -1;
  }

  printf ("Run \"demo4-creator -h\" for help.\n");
  printf ("Creating %d problems...\n", par.num);

  if (par.samurai)
    b = ConstructSamuraiBoard (par.blockWidth,
                               par.blockHeight,
                               par.diagonals);
  else
    b = ConstructCustomBoard (par.blockWidth,
                              par.blockHeight,
                              par.diagonals,
                              par.nGrids,
                              par.offsetsX,
                              par.offsetsY);

  if (!b)
    fprintf (stderr, "ERROR: Not enough free memory available\n");
  else
  {
    nCl = GetNumCells (b);
    cells = (int*) malloc (nCl*sizeof(int));
    symbols = (int*) malloc (nCl*sizeof(int));
    given = (char*) malloc ((nCl+1)*sizeof(char));

    if (!cells || !symbols || !given)
      fprintf (stderr, "ERROR: Not enough free memory available\n");
    else
    {
      serial = 0;

      if ((pf = fopen ("serial.txt", "rt")))
      {
        fscanf (pf, "%d", &serial);
        fclose (pf);

        if (serial<0 || serial>=MAX_NUM)
          serial = 0;
      }

      if ((pf = fopen ("serial.txt", "wt")))
      {
        fprintf (pf, "%d\n", serial+par.num);
        fclose (pf);
      }

      n = CreateProblems (b, &par,
                          cells, symbols, given,
                          serial);

      printf ("%d problems created.\n", n);
    }

    free (cells);
    free (symbols);
    free (given);

    DestroyBoard (b);
  }

  FreeParams (&par);

  SYSTEM_PAUSE;
  return 0;
}

/*////////////////////////////////////////////////////////////// */

int CreateProblems (t_board_p b, t_params * p,
                    int * cells, int * symbols,
                    char * given,
                    int serial)
{
  int np, trials, nSy, nCl;
  int i, j, k, n, c, s;
  e_state state;
  char name[50];
  FILE * pf;

  nSy = GetNumSymbols (b);
  nCl = GetNumCells (b);

  for (np=0; np<p->num; np++)
  {
    CleanBoard (b);
    n = trials = 0;

    do           /* First, generate a problem like in */
    {            /* demo3-minimal-creator */
      do
        c = my_random (nCl);
      while (GetNumPossValuesOfCell(b,c)==1 ||
             (p->restrictinters &&
              GetNumGroupsOfCell(b,c)>=
                      p->restrictinters));

      s = my_random (GetNumPossValuesOfCell(b,c));

      for (i=j=0; i<nSy; i++)
        if ((1<<i) & GetPossValuesOfCell(b,c))
        {
          if (j==s)
            break;
          else
            j ++;
        }

      s = i;

      SetSymbolInCell (b, s, c);

      Solve (b, 0, FALSE);

      state = GetState (b);

      if (state==impossible)
      {
        CleanBoard (b);

        if (trials<MAX_TRIALS)    /* Try to correct this some */
        {                         /* times before starting again */
          for (i=0; i<n; i++)
            SetSymbolInCell (b, symbols[i],
                                cells[i]);
          trials ++;
        }
        else
          n = trials = 0;   /* Start from scratch */

        continue;
      }

      symbols[n] = s;
      cells[n] = c;

      n ++;
    }
    while (state!=solved);

                          /* Now, try to remove */
    for (j=0; j<n;   )    /* unnecesary given values */
    {                     /* (No j++ here. See below) */
      CleanBoard (b);

      for (i=0; i<n; i++)        /* Set all but one */
        if (i!=j)
          SetSymbolInCell (b, symbols[i],
                              cells[i]);

      Solve (b, 0, FALSE);       /* Try to solve */

      if (GetState(b)==solved)   /* If it works, then that */
      {                          /* given value can be removed */
        n --;
        s = 1;

        for (i=j; i<n; i++)
        {
          symbols[i] = symbols[i+1];
          cells[i] = cells[i+1];
        }
      }
      else
        j ++;
    }

    CleanBoard (b);

    for (i=0; i<n; i++)
      SetSymbolInCell (b, symbols[i],
                          cells[i]);

    Solve (b, 0, FALSE);       /* Try solving again */
                               /* If it is difficult, leave it */
    if (!GetNumRulesNGt1(b))   /* If it's not so difficult */
    {                          /* make it easier at random */
      k = nCl/5 + my_random (
                    nCl/2-nCl/5 );   /* Choose a random number */
                                     /* of given values */
                                     /* 20%-50% of cells */
      for (i=0; i<nCl; i++)
        given[i] = 0;

      for (i=0; i<n; i++)
        given[cells[i]] = 1;

      if (n<k)                 /* Add given values up to */
        do                     /* the chosen number */
        {
          do
            c = my_random (nCl);
          while (given[c]);

          symbols[n] = GetSymbolOfCell (b, c);
          cells[n] = c;
          given[c] = 1;

          n ++;
        }
        while (n<k);
    }

    CleanBoard (b);

    for (i=0; i<n; i++)
      SetSymbolInCell (b, symbols[i],
                          cells[i]);

    Solve (b, 0, FALSE);

    if (GetNumRulesNGt2(b))
      sprintf (name, "%sN3x%02d-N2x%02d-%03d-%05d.txt",
                     p->diagonals ? "D-" : "",
                     GetNumRulesNGt2(b),
                     GetNumRulesN2(b),
                     nCl-n, np+serial);
    else if (GetNumRulesN2(b))
      sprintf (name, "%sN2x%02d-%03d-%05d.txt",
                     p->diagonals ? "D-" : "",
                     GetNumRulesN2(b),
                     nCl-n, np+serial);
    else
      sprintf (name, "%sN1-%03d-%05d.txt",
                     p->diagonals ? "D-" : "",
                     nCl-n, np+serial);

    pf = fopen (name, "wt");

    if (!pf)
    {
      perror (name);
      break;
    }

    CleanBoard (b);

    for (i=0; i<n; i++)
      SetSymbolInCell (b, symbols[i],
                          cells[i]);

    PrintBoard (b, pf, NULL, NULL);

    GetBoardRaw (b, given, NULL, NULL);

    fprintf (pf, "\n\n%d\n%s\n#\n%s\n",
                 n, name, given);

    for (i=0; i<40; i++)
      fputc ('\n', pf);

    Solve (b, 0, FALSE);

    PrintBoard (b, pf, NULL, NULL);

    GetBoardRaw (b, given, NULL, NULL);

    fprintf (pf, "\n@\n%s\n", given);

    fclose (pf);

    serial ++;

    if (serial>=MAX_NUM)
      serial = MAX_NUM;
  }

  return np;
}

/*////////////////////////////////////////////////////////////// */

int ParseParams (t_params * p, int argc, char *argv[])
{
  int i, j;
  char ok, wrong, incompatible, nomem;

  p->blockWidth = p->blockHeight = 3;   /* Default parameters */
  p->nGrids = 0;
  p->offsetsX = p->offsetsY = NULL;
  p->samurai = p->diagonals = FALSE;
  p->num = 10;
  p->restrictinters = 0;

  ok = TRUE;
  wrong = incompatible = nomem = FALSE;

  for (i=1; i<argc && ok; i++)  /* Parse parameters */

    if (!strcmp(argv[i],"-samurai"))  /* Samurai board */
    {
      if (p->nGrids)
      {
        ok = FALSE;
        incompatible = TRUE;
        break;
      }

      p->samurai = TRUE;
    }
    else if (!strcmp(argv[i],"-grids"))  /* Custom board */
    {
      if (p->samurai)
      {
        ok = FALSE;
        incompatible = TRUE;
        break;
      }

      i ++;

      if (i>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->nGrids = -1;
      p->nGrids = atoi (argv[i]);

      if (p->nGrids<2 || p->nGrids>MAX_GRIDS)
      {
        p->nGrids = 0;
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->offsetsX = (int*) malloc (p->nGrids * sizeof(int));
      p->offsetsY = (int*) malloc (p->nGrids * sizeof(int));

      if (!p->offsetsX || !p->offsetsY)
      {
        ok = FALSE;
        nomem = TRUE;
        break;
      }

      for (j=0; j<p->nGrids; j++, i+=2)
      {
        if (i+2>=argc)
        {
          ok = FALSE;
          wrong = TRUE;
          break;
        }

        p->offsetsX[j] = p->offsetsY[j] = MAX_DIST+1;

        p->offsetsX[j] = atoi (argv[i+1]);
        p->offsetsY[j] = atoi (argv[i+2]);

        if (p->offsetsX[j]<-MAX_DIST ||
            p->offsetsX[j]>MAX_DIST ||
            p->offsetsY[j]<-MAX_DIST ||
            p->offsetsY[j]>MAX_DIST)
        {
          ok = FALSE;
          wrong = TRUE;
          break;
        }
      }

      if (!ok)
        break;
    }
    else if (!strcmp(argv[i],"-blocksize"))  /* Block size */
    {
      i ++;

      if (i+1>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->blockWidth = p->blockHeight = -1;

      p->blockWidth = atoi (argv[i]);
      p->blockHeight = atoi (argv[i+1]);

      if (p->blockWidth < 2 ||
          p->blockHeight < 2 ||
          p->blockWidth*p->blockHeight > 30)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      i ++;
    }
    else if (!strcmp(argv[i],"-num"))  /* Number of problems */
    {                                  /* to create */
      i ++;

      if (i>=argc)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->num = -1;
      p->num = atoi (argv[i]);

      if (p->num < 1 ||
          p->num > MAX_NUM)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }
    }
    else if (!strcmp(argv[i],"-restrictinters")) /* Maximum */
    {                                            /* number of */
      i ++;                                    /* intersected */
                                             /* groups in cells */
      if (i>=argc)                         /* with given values */
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }

      p->restrictinters = -1;
      p->restrictinters = atoi (argv[i]);

      if (p->restrictinters < 0)
      {
        ok = FALSE;
        wrong = TRUE;
        break;
      }
    }
    else if (!strcmp(argv[i],"-diagonals"))  /* Diagonals */
    {
      p->diagonals = TRUE;
    }
    else if (!strcmp(argv[i],"-h"))  /* Help */
    {
      printf ("\n\tUSAGE:\n\tdemo4-creator "
              "[parameters]\n\n"
              "\tPOSSIBLE PARAMETERS:\n"
              "\t  -h\t(show this help)\n"
              "\t  -samurai\n"
              "\t  -grids n x1 y1 x2 y2... xn yn\n"
              "\t  -blocksize width height\n"
              "\t  -diagonals\n"
              "\t  -num n\n"
              "\t  -restrictinters numinters\n"
              "\n");

      FreeParams (p);
      return -1;
    }
    else
    {
      ok = FALSE;
      wrong = TRUE;
    }

  if (ok)
    return 0;

  if (wrong)
    fprintf (stderr, "ERROR: Wrong parameters\n");
  else if (incompatible)
    fprintf (stderr, "ERROR: Incompatible parameters\n");
  else if (nomem)
    fprintf (stderr, "ERROR: Not enough free memory available\n");

  FreeParams (p);
  return -1;
}

/*////////////////////////////////////////////////////////////// */

void FreeParams (t_params * p)
{
  if (p->offsetsX)
  {
    free (p->offsetsX);
    p->offsetsX = NULL;
  }

  if (p->offsetsY)
  {
    free (p->offsetsY);
    p->offsetsY = NULL;
  }
}

/*////////////////////////////////////////////////////////////// */

int my_random (int num)
{
  double d;
  int r;

  d = rand () / (double) RAND_MAX;
  r = (int) (d*num);

  return r<0 ? 0 : r>=num ? num-1 : r;
}

